# KatapultAPI

katapultAPI is a Python module designed to handle requests made to the Katapult API.

## Installation & Setup

```bash
pip install katapult-api
```

To develop and iterate upon the Katapult API module first clone the repo:

```shell
git clone https://github.com/TechServ-Consulting-Training-Ltd/katapult-api
```

then run the following commands from the project root:

```shell
python -m venv ./.venv
./.venv/Scripts/activate
pip install -r requirements-dev.txt
pre-commit install
```

## Usage

[//]: # (todo maybe a usage of this in a mainly sychronous process)

[//]: # (todo probably short explaination to expand below usage)

### Getting a Job list and Job Details

```python
import logging.config
import asyncio

# Not required but the client itself has logging functionality to help track requests.
logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.DEBUG,
    format='[%(levelname)s] %(asctime)s - %(name)s - %(message)s'
)

from katapult_api.katapult import Katapult
from katapult_api.utils import task_handler


async def main():
    # Sets up the aiohttp session and closes it after finishing.
    async with Katapult(api_key="YOURAPIKEY") as katapult:
        # Returns a Response object that contains attributes such as return status, content, and even a json() method.
        job_lists = await katapult.job_lists()

        # Here we build 20 payloads that will be sent out at the same time.
        job_ids = [{"job_id": key} for key in list(job_lists.json().keys())[:20]]

        # Not required but makes setting up async tasks easy, pass in a job name and a list of dicts that contain
        # expected params.
        response = await task_handler(katapult.job, job_ids)
    return response


if __name__ == '__main__':
    logger.info(f"Getting jobs...")
    jobs = asyncio.run(main())
    logger.info(f"Finished, got {len(jobs)} jobs.")
```

### Creating and Updating a Node

```python
import asyncio

# Possible logging...

from katapult_api.katapult import Katapult


async def main():
    async with Katapult(api_key="YOURAPIKEY") as katapult:
        # Get job id from job_lists endpoint...
        job_id = "-ABC123"

        # Create a node by passing job id, lat/long, and a flat dict of attributes to write to the node.
        response = await katapult.create_node(job_id, 30.0817, -97.843, {"OID": 123, "my_custom_attr": "howdy partner"})
    return response


if __name__ == '__main__':
    results = asyncio.run(main())
```

The above should result in something like this:

![img.png](docs/attachments/create_node_example.png)

Similarly, a node can be updated by passing both job and node ids:
```python
import asyncio

# Possible logging...

from katapult_api.katapult import Katapult


async def main():
  async with Katapult(api_key="YOURAPIKEY") as katapult:
    # Get job id from job_lists endpoint...
    job_id = "-ABC123"

    # Get node id from parsing job from job endpoint...
    node_id = "-EFG456"

    # Similar to the create node method but needs a nodeid. Note that lat/long and attributes are
    # technically optional; in this example lat/long are not included.
    response = await katapult.update_node(jobid=job_id, nodeid=node_id, attributes={"my_fav_pokemon": "jigglypuff"})

  return response


if __name__ == '__main__':
  results = asyncio.run(main())
```

### KatapultNice Usage

```python
import logging.config
import asyncio

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)

from katapult_api.katapult_nice import KatapultNice
from katapult_api.utils import task_handler


async def main(api_key: str):
    async with KatapultNice(api_key=api_key) as nice:
        job_lists = await nice.job_lists()
        job_ids = [{"job_id": key} for key in list(job_lists.json().keys())[:20]]
        response = await task_handler(nice.job_nodes, job_ids)
    return response


if __name__ == '__main__':
    api_key = "api_key"
    logger.info(f"Getting jobs...")
    jobs = asyncio.run(main(api_key))
    logger.info(f"Finished, got {len(jobs)} jobs.")

```

## Contributing

Pull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.

Please make sure to update tests as appropriate.

### Building

For building and updating the PYPI project found here: https://pypi.org/project/katapult-api/, some prerequisites are
required.

- You will also need a PYPI account and to be named a collaborator for this project with a maintain role.
- In your home directory (windows: ```C:/Users/username```, linux: ```/home/username``` ) create a ```.pypirc``` file with the
  following structure

```pypirc
[distutils]
index-servers = pypi

[pypi]
repository = https://upload.pypi.org/legacy/
username = __token__
password = pypi-<YOUR API KEY>
```

- Ensure you have both twine and build installed (should be installed from ```requirements-dev.txt```).

Once you have the above set-up you are ready to build and update the PYPI project.

To build, first update the version number in the ```pyproject.toml```. Note that if you do not do this, twine will throw an
error:

```toml
...

[project]
name = "katapult-api"
version = "1.0.0" # update this

...
```

Next build the project. This should create a ```dist``` folder:

```shell
python -m build
```

Next run the following command. This will update the project on PYPI:
```shell
twine upload dist/*
```

## Resources

- https://documenter.getpostman.com/view/9081167/SVtVV93W: Katapultpro API V1
- https://github.com/KatapultDevelopment/katapult-pro-api-documentation?tab=readme-ov-file: Katapultpro Github

[//]: # (todo maybe a list of currently supported methods?)
